@@ -14,7 +14,9 @@ from collections import namedtuple
 import logging
 import math
 import os
+import six
 from six.moves import queue
+import sys
 
 from awscli.customizations.s3.constants import MULTI_THRESHOLD, CHUNKSIZE, \
     NUM_THREADS, MAX_UPLOAD_SIZE, MAX_QUEUE_SIZE
@@ -36,6 +38,8 @@ class S3Handler(object):
     class pull tasks from to complete.
     """
     MAX_IO_QUEUE_SIZE = 20
+    MAX_EXECUTOR_QUEUE_SIZE = MAX_QUEUE_SIZE
+    EXECUTOR_NUM_THREADS = NUM_THREADS
 
     def __init__(self, session, params, result_queue=None,
                  multi_threshold=MULTI_THRESHOLD, chunksize=CHUNKSIZE):
@@ -53,7 +57,9 @@ class S3Handler(object):
                        'content_type': None, 'cache_control': None,
                        'content_disposition': None, 'content_encoding': None,
                        'content_language': None, 'expires': None,
-                       'grants': None}
+                       'grants': None, 'only_show_errors': False,
+                       'is_stream': False, 'paths_type': None,
+                       'expected_size': None}
         self.params['region'] = params['region']
         for key in self.params.keys():
             if key in params:
@@ -61,8 +67,11 @@ class S3Handler(object):
         self.multi_threshold = multi_threshold
         self.chunksize = chunksize
         self.executor = Executor(
-            num_threads=NUM_THREADS, result_queue=self.result_queue,
-            quiet=self.params['quiet'], max_queue_size=MAX_QUEUE_SIZE,
+            num_threads=self.EXECUTOR_NUM_THREADS,
+            result_queue=self.result_queue,
+            quiet=self.params['quiet'],
+            only_show_errors=self.params['only_show_errors'],
+            max_queue_size=self.MAX_EXECUTOR_QUEUE_SIZE,
             write_queue=self.write_queue
         )
         self._multipart_uploads = []
@@ -234,12 +243,11 @@ class S3Handler(object):
         create_file_task = tasks.CreateLocalFileTask(context=context,
                                                      filename=filename)
         self.executor.submit(create_file_task)
-        for i in range(num_downloads):
-            task = tasks.DownloadPartTask(
-                part_number=i, chunk_size=chunksize,
-                result_queue=self.result_queue, service=filename.service,
-                filename=filename, context=context, io_queue=self.write_queue)
-            self.executor.submit(task)
+        self._do_enqueue_range_download_tasks(
+            filename=filename, chunksize=chunksize,
+            num_downloads=num_downloads, context=context,
+            remove_remote_file=remove_remote_file
+        )
         complete_file_task = tasks.CompleteDownloadTask(
             context=context, filename=filename, result_queue=self.result_queue,
             params=self.params, io_queue=self.write_queue)
@@ -251,6 +259,16 @@ class S3Handler(object):
             self.executor.submit(remove_task)
         return num_downloads
 
+    def _do_enqueue_range_download_tasks(self, filename, chunksize,
+                                         num_downloads, context,
+                                         remove_remote_file=False):
+        for i in range(num_downloads):
+            task = tasks.DownloadPartTask(
+                part_number=i, chunk_size=chunksize,
+                result_queue=self.result_queue, service=filename.service,
+                filename=filename, context=context, io_queue=self.write_queue)
+            self.executor.submit(task)
+
     def _enqueue_multipart_upload_tasks(self, filename,
                                         remove_local_file=False):
         # First we need to create a CreateMultipartUpload task,
@@ -295,14 +313,27 @@ class S3Handler(object):
         self.executor.submit(create_multipart_upload_task)
         return upload_context
 
-    def _enqueue_upload_tasks(self, num_uploads, chunksize, upload_context, filename,
-                              task_class):
+    def _enqueue_upload_tasks(self, num_uploads, chunksize, upload_context,
+                              filename, task_class):
         for i in range(1, (num_uploads + 1)):
-            task = task_class(
-                part_number=i, chunk_size=chunksize,
-                result_queue=self.result_queue, upload_context=upload_context,
-                filename=filename)
-            self.executor.submit(task)
+            self._enqueue_upload_single_part_task(
+                part_number=i,
+                chunk_size=chunksize,
+                upload_context=upload_context,
+                filename=filename,
+                task_class=task_class
+            )
+
+    def _enqueue_upload_single_part_task(self, part_number, chunk_size,
+                                         upload_context, filename, task_class,
+                                         payload=None):
+        kwargs = {'part_number': part_number, 'chunk_size': chunk_size,
+                  'result_queue': self.result_queue,
+                  'upload_context': upload_context, 'filename': filename}
+        if payload:
+            kwargs['payload'] = payload
+        task = task_class(**kwargs)
+        self.executor.submit(task)
 
     def _enqueue_upload_end_task(self, filename, upload_context):
         complete_multipart_upload_task = tasks.CompleteMultipartUploadTask(
@@ -311,3 +342,157 @@ class S3Handler(object):
         self.executor.submit(complete_multipart_upload_task)
         self._multipart_uploads.append((upload_context, filename))
 
+
+class S3StreamHandler(S3Handler):
+    """
+    This class is an alternative ``S3Handler`` to be used when the operation
+    involves a stream since the logic is different when uploading and
+    downloading streams.
+    """
+
+    # This ensures that the number of multipart chunks waiting in the
+    # executor queue and in the threads is limited.
+    MAX_EXECUTOR_QUEUE_SIZE = 2
+    EXECUTOR_NUM_THREADS = 6
+
+    def _enqueue_tasks(self, files):
+        total_files = 0
+        total_parts = 0
+        for filename in files:
+            num_uploads = 1
+            # If uploading stream, it is required to read from the stream
+            # to determine if the stream needs to be multipart uploaded.
+            payload = None
+            if filename.operation_name == 'upload':
+                payload, is_multipart_task = \
+                    self._pull_from_stream(self.multi_threshold)
+            else:
+                # Set the file size for the ``FileInfo`` object since
+                # streams do not use a ``FileGenerator`` that usually
+                # determines the size.
+                filename.set_size_from_s3()
+                is_multipart_task = self._is_multipart_task(filename)
+            if is_multipart_task and not self.params['dryrun']:
+                # If we're in dryrun mode, then we don't need the
+                # real multipart tasks.  We can just use a BasicTask
+                # in the else clause below, which will print out the
+                # fact that it's transferring a file rather than
+                # the specific part tasks required to perform the
+                # transfer.
+                num_uploads = self._enqueue_multipart_tasks(filename, payload)
+            else:
+                task = tasks.BasicTask(
+                    session=self.session, filename=filename,
+                    parameters=self.params,
+                    result_queue=self.result_queue,
+                    payload=payload)
+                self.executor.submit(task)
+            total_files += 1
+            total_parts += num_uploads
+        return total_files, total_parts
+
+    def _pull_from_stream(self, amount_requested):
+        """
+        This function pulls data from stdin until it hits the amount
+        requested or there is no more left to pull in from stdin.  The
+        function wraps the data into a ``BytesIO`` object that is returned
+        along with a boolean telling whether the amount requested is
+        the amount returned.
+        """
+        stream_filein = sys.stdin
+        if six.PY3:
+            stream_filein = sys.stdin.buffer
+        payload = stream_filein.read(amount_requested)
+        payload_file = six.BytesIO(payload)
+        return payload_file, len(payload) == amount_requested
+
+    def _enqueue_multipart_tasks(self, filename, payload=None):
+        num_uploads = 1
+        if filename.operation_name == 'upload':
+            num_uploads = self._enqueue_multipart_upload_tasks(filename,
+                                                               payload=payload)
+        elif filename.operation_name == 'download':
+            num_uploads = self._enqueue_range_download_tasks(filename)
+        return num_uploads
+
+    def _enqueue_range_download_tasks(self, filename, remove_remote_file=False):
+
+        # Create the context for the multipart download.
+        chunksize = find_chunksize(filename.size, self.chunksize)
+        num_downloads = int(filename.size / chunksize)
+        context = tasks.MultipartDownloadContext(num_downloads)
+
+        # No file is needed for downloading a stream.  So just announce
+        # that it has been made since it is required for the context to
+        # begin downloading.
+        context.announce_file_created()
+
+        # Submit download part tasks to the executor.
+        self._do_enqueue_range_download_tasks(
+            filename=filename, chunksize=chunksize,
+            num_downloads=num_downloads, context=context,
+            remove_remote_file=remove_remote_file
+        )
+        return num_downloads
+
+    def _enqueue_multipart_upload_tasks(self, filename, payload=None):
+        # First we need to create a CreateMultipartUpload task,
+        # then create UploadTask objects for each of the parts.
+        # And finally enqueue a CompleteMultipartUploadTask.
+
+        chunksize = self.chunksize
+        # Determine an appropriate chunksize if given an expected size.
+        if self.params['expected_size']:
+            chunksize = find_chunksize(int(self.params['expected_size']),
+                                       self.chunksize)
+        num_uploads = '...'
+
+        # Submit a task to begin the multipart upload.
+        upload_context = self._enqueue_upload_start_task(
+            chunksize, num_uploads, filename)
+
+        # Now submit a task to upload the initial chunk of data pulled
+        # from the stream that was used to determine if a multipart upload
+        # was needed.
+        self._enqueue_upload_single_part_task(
+            part_number=1, chunk_size=chunksize,
+            upload_context=upload_context, filename=filename,
+            task_class=tasks.UploadPartTask, payload=payload
+        )
+
+        # Submit tasks to upload the rest of the chunks of the data coming in
+        # from standard input.
+        num_uploads = self._enqueue_upload_tasks(
+            num_uploads, chunksize, upload_context,
+            filename, tasks.UploadPartTask
+        )
+
+        # Submit a task to notify the multipart upload is complete.
+        self._enqueue_upload_end_task(filename, upload_context)
+
+        return num_uploads
+
+    def _enqueue_upload_tasks(self, num_uploads, chunksize, upload_context,
+                              filename, task_class):
+        # The previous upload occured right after the multipart
+        # upload started for a stream.
+        num_uploads = 1
+        while True:
+            # Pull more data from standard input.
+            payload, is_remaining = self._pull_from_stream(chunksize)
+            # Submit an upload part task for the recently pulled data.
+            self._enqueue_upload_single_part_task(
+                part_number=num_uploads+1,
+                chunk_size=chunksize,
+                upload_context=upload_context,
+                filename=filename,
+                task_class=task_class,
+                payload=payload
+            )
+            num_uploads += 1 
+            if not is_remaining:
+                break
+        # Once there is no more data left, announce to the context how
+        # many parts are being uploaded so it knows when it can quit.
+        upload_context.announce_total_parts(num_uploads)
+        return num_uploads
